module Datentypen where 
import Data.List
import Data.Either
import Debug.Trace

type TermNT    = String
type ContextNT = String
type FunSym    = String
type Variable  = String

----Für die Eingabe-Grammatik kann man entweder mit dem Datentyp STGProd oder mit dem Datentyp STG_Prod arbeiten. Wenn man STGProd benutzt kann man
----die Regeln der STG-Grammatik direkt an den Konstruktoren sehen. Dies erleichtert die Eingabe für Benutzer die mit den Grammatik-Regeln nicht vertraut sind.
----Bei STG_Prod muss man die Produktionen so entwerfen, dass sie den Regeln schon genügen. Hierbei ist es von Nützen wenn man die Regeln schon im Kopf hat, was einem viel ----Arbeit erspart, denn sonst muss man immer nachschauen welche Produktionen erlaubt sind und welche nicht.
----------------------------------------





data STGProd       =  AfA TermNT FunSym [TermNT]
                    | ACA TermNT ContextNT TermNT
                    | CLoch ContextNT
                    | CCC  ContextNT ContextNT ContextNT
                    | CfAC ContextNT FunSym [TermNT] ContextNT [TermNT]
                    | Lam  TermNT TermNT
                    | Ax TermNT Variable
 deriving(Eq,Read)

instance Show(STGProd) where
 show (AfA a f []) = a ++ " --> " ++ f
 show (AfA a f as) = a ++ " --> " ++ f ++ "(" ++ concat (intersperse "," as) ++ ")"
 show (ACA a1 c a2)  = a1 ++ " --> " ++ c ++ " " ++ a2
 show (CLoch c)    = c ++ " --> []"
 show (CCC c1 c2 c3)    = c1 ++ " --> "  ++ c2 ++ " " ++ c3
 show (CfAC c1 f as1  c2 as2)    = if as1 == []
                                   then
                                         if as2 == []
                                         then
                                         c1 ++ " --> " ++ f ++ "(" ++ c2 ++ ")"
                                         else
                                         c1 ++ " --> " ++ f ++ "(" ++ c2 ++ "," ++ concat (intersperse "," as2) ++ ")"
                                   else
                                         if as2 == []
                                         then
                                         c1 ++ " --> " ++ f ++ "(" ++ concat (intersperse "," as1) ++ "," ++ c2 ++ ")"
                                         else 
                                         c1 ++ " --> " ++ f ++ "(" ++ concat (intersperse "," as1) ++ "," ++ c2 ++ "," ++ concat (intersperse "," as2) ++ ")"
 show (Lam a1 a2) = a1 ++ " --> " ++ a2
 show (Ax a x) = a ++ " --> " ++ x

newtype STG = STG [STGProd]
 deriving(Eq,Read)

instance Show(STG) where
 show (STG as) = unlines (map show as)

---------------------------------------
checkSTG :: STG -> Bool
checkSTG (stg)
 | checkSets (getSets stg) = checkFnArities stg []
 | otherwise = False

checkSets :: ([TermNT],[ContextNT],[FunSym],[Variable],[Either TermNT ContextNT]) -> Bool
checkSets (termNT,contextNT,funsym,vars,defnts) = 
 let tntStrings = nub $ termNT
     cntStrings = nub $ contextNT
     funStrings = nub $ funsym
     varStrings = nub $ vars
     defTNTs =   lefts defnts
     defCNTs =   rights defnts
     test = [ ((all null [as `intersect` bs | (as,bs)  <- [(tntStrings,cntStrings), (tntStrings,varStrings), (tntStrings,funStrings),                                                       (cntStrings,varStrings), (cntStrings,funStrings),
                                                        (varStrings,funStrings)]])
                 , "Namen der syntaktischen Kategorien \"uberlappen"
                 )
               , ((all (\x -> x `elem` defTNTs) tntStrings)
                 ,
                   "Es gibt undefinierte Term-Nichterminale"
                 )
               , ((all (\x -> x `elem` defCNTs) cntStrings)
                 ,
                   "Es gibt undefinierte Context-Nichterminale"
                 )
               , (length defTNTs == length (nub defTNTs)
                 ,
                   "Es gibt mehr als eine Produktion f\"ur ein Term-Nichterminal"
                 )
               , (length defCNTs == length (nub defCNTs)
                 ,
                   "Es gibt mehr als eine Produktion f\"ur ein ContextNichterminal"
                 )
               ]
 in mkChecks test
  where mkChecks [] = True
        mkChecks ((x,emess):xs) = if x then mkChecks xs else error emess
       

getSets :: STG -> ([TermNT],[ContextNT],[FunSym],[Variable],[Either TermNT ContextNT])
getSets (STG []) = ([],[],[],[],[])
getSets (STG (x:xs))  = 
 let rekset@(termNT,contextNT,funsym,vars,defnts) = getSets (STG xs)
 in case x of 
    AfA a f as            -> (((a:as) ++ termNT), contextNT, f:funsym,vars,(Left a):defnts)
    ACA a1 c a2           -> (a1:a2:termNT, c:contextNT, funsym,vars,(Left a1:defnts))
    CLoch c               -> (termNT, c:contextNT, funsym,vars,(Right c):defnts)
    CCC c1 c2 c3          -> (termNT, c1:c2:c3:contextNT, funsym,vars,(Right c1):defnts)
    CfAC c1 f as1 c2 as2  -> (as1++as2++termNT, c1:c2:contextNT, f:funsym,vars,(Right c1):defnts)
    Lam a1 a2             -> (a1:a2:termNT,contextNT,funsym,vars,(Left a1):defnts)
    Ax a x                -> (a:termNT,contextNT,funsym,x:vars,(Left a):defnts)


 
---------------------------------------

checkFnArities :: STG -> [(FunSym, Int)] -> Bool
checkFnArities (STG [])     arities = True
checkFnArities (STG ((AfA a f as):xs)) arities = 
  case lookup f arities of
     Nothing -> let new_arities = (f,length as):arities
                in checkFnArities (STG xs) new_arities
     Just x -> if length as == x then checkFnArities (STG xs) arities 
              else error $ "Funktionssymbol " ++ show f ++ "mit Stelligkeit " ++ show x ++ " und " ++ show (length as) ++ " verwendet!"

checkFnArities (STG ((CfAC c1 f as1 c2 as2):xs)) arities = 
  case lookup f arities of
     Nothing -> let new_arities = (f,1+(length as1)+(length as2)):arities
                in checkFnArities (STG xs) new_arities
     Just x ->  if 1+(length as1) + (length as2)  == x then checkFnArities (STG xs) arities 
              else error $ "Funktionssymbol " ++ show f ++ "mit Stelligkeit " ++ show x ++ " und " ++ show (1+(length as1) + (length as2)) ++ " verwendet!"

checkFnArities (STG (x:xs)) arities = checkFnArities (STG xs) arities

---------------------------------------



convertSTGtoSTG_Prod :: STG -> [STG_Prod]
convertSTGtoSTG_Prod (STG stg) 
 | checkSTG (STG stg) =  map convertProd stg
 | otherwise = error "Check for STG failed"

convertProd (AfA a f [])   = STG_Prod (TN a) [K f]
convertProd (AfA a f as)   = STG_Prod (TN a) ((F f):(map TN as))
convertProd (ACA a1 c a2)  = STG_Prod (TN a1) [CN c,TN a2]
convertProd (CLoch c)      = STG_Prod (CN c) [Loch]
convertProd (CCC c1 c2 c3) = STG_Prod (CN c1) [CN c2, CN c3]
convertProd (CfAC c1 f as1  c2 as2)    =  STG_Prod (CN c1) ((F f:(map TN as1)) ++ ((CN c2):(map TN as2)))
convertProd (Lam a1 a2)    = STG_Prod (TN a1) [TN a2]
convertProd (Ax a x)       = STG_Prod (TN a) [X x]

beispielSTG = STG [AfA "A" "f" ["A1","A2"]
               ,Ax "A1" "x"
               ,AfA "A2" "g" []
               ]


stgprod_in_stg :: [STG_Prod] -> [STGProd] -> STG
stgprod_in_stg []     prods = (STG prods)
stgprod_in_stg (x:xs) prods = case x of
                                   (STG_Prod (TN a) [K f])          -> stgprod_in_stg xs (prods++[(AfA a f [])])
                                   (STG_Prod (TN a) [TN b])         -> stgprod_in_stg xs (prods++[(Lam a b)])
                                   (STG_Prod (TN a1) [CN c,TN a2])  -> stgprod_in_stg xs (prods++[(ACA a1 c a2)])
                                   (STG_Prod (TN a) [X x])          -> stgprod_in_stg xs (prods++[(Ax a x)])
                                   (STG_Prod (TN a) ((F f):ys))     -> stgprod_in_stg xs (prods++[(AfA a f (nimmStr ys []))])
                                   (STG_Prod (CN a) [Loch])         -> stgprod_in_stg xs (prods++[(CLoch a)])
                                   (STG_Prod (CN c1) [CN c2,CN c3]) -> stgprod_in_stg xs (prods++[(CCC c1 c2 c3)])
                                   (STG_Prod (CN c1) ((F f):ys))    -> let t = (teile_in_3 ys [] "" []) in
                                                                       stgprod_in_stg xs (prods++[(CfAC c1 f (first_d t) (second_d t) (third_d t))])

nimmStr :: [STG_Pre_Symbol] -> [String] -> [String]
nimmstr []     s = s
nimmStr (x:xs) s = if xs==[]
                   then
                   case x of
                        (TN a)    -> (s++[a])
                        otherwise -> error "1.TN A --> f (liste) , liste darf nur TN's enthalten !!! (Module Datentypen,Funktion nimmStr)"
                   else
                   case x of
                        (TN a)    -> nimmStr xs (s++[a])
                        otherwise -> error "2.TN A --> f (liste) , liste darf nur TN's enthalten !!! (Module Datentypen,Funktion nimmStr)"

first_d::([String],String,[String]) -> [String]
first_d  (a,b,c) = a
second_d::([String],String,[String]) -> String
second_d (a,b,c) = b
third_d::([String],String,[String]) -> [String]
third_d  (a,b,c) = c

teile_in_3 :: [STG_Pre_Symbol] -> [String] -> String -> [String] -> ([String],String,[String])
teile_in_3 []     as1 c as2 = (as1,c,as2)
teile_in_3 (x:xs) as1 c as2 = case x of
                                   (TN a) -> if c==""
                                             then teile_in_3 xs (as1++[a]) c as2
                                             else teile_in_3 xs as1 c (as2++[a])
                                   (CN a) -> teile_in_3 xs as1 (c++a) as2

------------------

--------------------------------STG_in_SCFG
data STG_Pre_Symbol = TN String      --TN sind die Term-Nonterminale der STG
                    | CN String      --CN sind die Kontext-Nonterminale der STG
                    | Loch           --ist das Loch in einem Kontext
                    | P String       --aus einem TN der STG wird nach der Preorder-Traversierung ein P in der ProdPreOrder-Grammatik
                    | L String       --die CN der STG werden bei der Preorder-Traversierung in die linke und rechte Hälfte vom Loch aufgeteilt.
                                     --L stellt die linke Hälfte vom Loch des Kontexts in der ProdPreOrder-Grammatik dar
                    | R String       --R stellt die rechte Hälfte vom Loch des Kontexts in der ProdPreOrder-Grammatik dar
                    | F String       --F sind Funktionssymbole mit Stelligkeit >0 und werden sowohl in der STG, als auch in der ProdPreOrder-Grammatik verwendet
                    | X String       --X sind Variablen und werden sowohl in der STG, als auch in der ProdPreOrder-Grammatik verwendet
                    | K String       --K sind Konstanten (Funktionssymbole mit Stelligkeit 0) und werden sowohl in der STG, als auch in der ProdPreOrder-Grammatik verwendet.
                    | Laenge Integer --repräsentiert die Laenge des Wortes das durch ein Nonterminal gebildet wird. Sie wird sowohl in der STG, als auch in der
                                     --ProdPreOrder-Grammatik verwendet.
                    | Lambda         --aus einem Loch in der STG wird nach der Preorder-Traversierung ein Lambda in der ProdPreOrder-Grammatik

 deriving (Eq,Ord,Show,Read)

data STG_Prod = STG_Prod (STG_Pre_Symbol) [STG_Pre_Symbol] deriving (Eq,Ord,Read)

instance Show(STG_Prod) where
 show (STG_Prod (TN a) as) = case as of
                                  ((F c):xs) -> "TN " ++ a ++ " --> " ++ " F "++ c ++" (" ++ (intersperse1 xs "") ++ ")\n"
                                  otherwise  -> "TN " ++ a ++ " --> " ++ (intersperse1 as "") ++ "\n"

 show (STG_Prod (CN a) as) = case as of
                                  ((F c):xs) -> "CN " ++ a ++ " --> " ++ " F "++ c ++" (" ++ (intersperse1 xs "") ++ ")\n"
                                  [Loch]     -> "CN " ++ a ++ " --> " ++ "[]\n"
                                  otherwise  -> "CN " ++ a ++ " --> " ++ (intersperse1 as "") ++ "\n"


intersperse1:: [STG_Pre_Symbol] -> String -> String
intersperse1 []     s = s
intersperse1 (x:xs) s = if xs==[]
                        then
                            case x of
                                 TN b -> intersperse1 xs (s++("TN "++b))
                                 CN b -> intersperse1 xs (s++("CN "++b))
                                 F  b -> intersperse1 xs (s++("F "++b))
                                 K  b -> intersperse1 xs (s++("K "++b))
                                 X  b -> intersperse1 xs (s++("X "++b))

                        else
                            case x of
                                 TN b -> intersperse1 xs (s++("TN "++b++","))
                                 CN b -> intersperse1 xs (s++("CN "++b++","))
                                 F  b -> intersperse1 xs (s++("F "++b++","))
                                 K  b -> intersperse1 xs (s++("K "++b++","))
                                 X  b -> intersperse1 xs (s++("X "++b++","))


data ProdPre = ProdPreOrder (STG_Pre_Symbol) [STG_Pre_Symbol] deriving (Eq,Ord,Show,Read)


------------Position
data Position = Pos (Pos_Symbol) [Pos_Symbol] deriving (Eq,Ord,Show)
data Pos_Symbol = PT String     --PT sind die Positions-Nonterminale der Positionsgrammatik. Aus einem TN in der STG wird ein PT in der Positionsgrammatik.
                | H String      --H sind die Kontext-Nonterminale der Positionsgrammatik. Aus einem CN in der STG wird ein H in der Positionsgrammatik.
                | Z Integer     --geben den Weg zu dem Unterterm an, in dem sich die zu unifizierenden Terme unterscheiden. (sind Zahlen mit deren Hilfe der Weg zu einem
                                --Unterterm gefunden wird)
                | Laeng Integer --repräsentiert die Laenge des Wortes das durch ein Nonterminal gebildet wird. (gibt an in welcher Tiefe der zu unifizierende Term sich 
                                --befindet)
 deriving (Eq,Ord,Show,Read)


-------------------------Plandowski
data Symbol  = T String  --Terminalsymbol
             | N String  --Nonterminal
             | I Integer --Länge des Wortes das durch ein Nonterminal gebildet wird
 deriving (Eq,Ord,Show,Read)

data ProdGen = ProdSymSym (Symbol) (Symbol,Symbol)
             | ProdIntSym (Symbol) (Integer, Symbol)
             | ProdSymInt (Symbol) (Symbol,Integer)
             | ProdInt (Symbol) (Integer) deriving (Eq,Ord,Show,Read)

data Prod  = Produktion (Symbol) [Symbol]  deriving (Eq,Ord,Show,Read)
