module PlandowskiSTG_Prod.ChomskyProd where

import Datentypen


chomsky :: [Prod] -> Integer -> [Prod]
chomsky prod zaehler = let chs = chomsky_sub prod zaehler in
                         (first chs ++ second chs)

chomsky_sub :: [Prod] -> Integer -> ([Prod],[Prod],[Prod])
chomsky_sub prod zaehler = let cnf2 = (cnf_reg_2 prod) in
                                 if thrd cnf2 == [] then 
                                                    cnf_reg_3 cnf2 [] zaehler
                                                    else
                                                    cnf_reg_2_5 cnf2 [] [] zaehler



first :: ([a],[a],[a])->[a]
first (x,y,z) = x

second :: ([a],[a],[a])->[a]
second (x,y,z) = y

thrd :: ([a],[a],[a])->[a]
thrd (x,y,z) = z


---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------
--3.SCHRITT TRANSFORMATION DER PLANDOWSKI SCFG in CNF-PLANDOWSKI SCFG -----------------
---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------
--0   Regel der CNF: Ersetze alle Regeln der Form : N a -> N b und N b -> T x (oder N a -> N b und N b -> N c usw.), durch eine neue Regel N a -> T x (die Regel N a -> N b wird nicht mehr benötigt)
--1.  Regel der CNF: Regeln N a -> T b werden übernommen => snd (split_in_tupel_prod (umwandeln( beispiel1)))  kommt in CNF Menge
--2.  Regel der CNF: Für alle Terminalsymbole aus der Liste  "tprods" (die auf der rechten Seite  alleine stehen), werden in "t_n_prods" eingesetzt.
--2.3 Regel der CNF: Ersetze alle Regeln der Form : N a -> N b und N b -> T x (oder N a -> N b und N b -> N c usw.), durch eine neue Regel N a -> T x (die Regel N a -> N b wird nicht mehr benötigt)
--2.5 Regel der CNF: Für alle Terminalsymbole, die auf der rechten Seite nicht alleine stehen (t_n_prods) zb. (N a -> Tb, N d ,N c), werden 
--                   Regeln N "C_Tb" -> T b eingeführt und in (N a -> N a, T b ,N c) eingesetzt => (N a -> N a, N "C_Tb" ,N c)
--3.  Regel der CNF: Die restlichen Regeln haben die Form N a -> N y1....N yn. Wenn n/=2 werden weitere Nichtterminale eingeführt,
--                   sodass N a -> N y1 N z1 bis N zn-2 -> N yn-1 N yn. Die Ausgabe ist eine Liste der  Epsilon/Lambda freien Produktionen...

 








--0 Regel der CNF: Ersetze alle Regeln der Form : N a -> N b und N b -> T x (oder N a -> N b und N b -> N c usw.), durch eine neue Regel N a -> T x (die Regel N a -> N b wird nicht mehr benötigt)
cnf_reg_0 :: [Prod] -> [Prod]
cnf_reg_0 prod = let stp = (split_in_tupel_prod1 prod) in
                 durchlaufe_lambda_rules (fst stp) (snd stp) []

durchlaufe_lambda_rules :: [Prod] -> [Prod] -> [Prod] -> [Prod]
durchlaufe_lambda_rules prod [] letzte = if letzte==[] then prod else letzte
durchlaufe_lambda_rules prod (y:lambdaprod) letzte = let slp = (subst_lambda_rules_in_prods prod y)
                                                         sll = (subst_lambda_rules_in_lambdas lambdaprod y) in 
                                                     durchlaufe_lambda_rules slp sll (slp++sll)

subst_lambda_rules_in_prods :: [Prod] -> Prod -> [Prod]
subst_lambda_rules_in_prods [] x = [] 
subst_lambda_rules_in_prods ((Produktion x xs):ys) (Produktion (p) [k]) = let slp = subst_lambda_rules_in_prods ys in
                                                                          if x==k 
                                                                          then 
                                                                          (Produktion x xs) : (Produktion (p) xs) : slp (Produktion (p) [k]) 
                                                                          else 
                                                                          (Produktion x xs) : slp (Produktion (p) [k])


subst_lambda_rules_in_lambdas :: [Prod] -> Prod -> [Prod]
subst_lambda_rules_in_lambdas [] x = []
subst_lambda_rules_in_lambdas ((Produktion (x) [z]):xs) (Produktion (p) [k]) = let sll = subst_lambda_rules_in_lambdas xs in
                                                                               if x==k 
                                                                               then
                                                                               (Produktion (x) [z]) : (Produktion (p) [z]) : sll (Produktion (p) [k])
                                                                               else 
                                                                               (Produktion (x) [z]) : sll (Produktion (p) [k])


ersetze_prod :: Symbol -> Symbol -> [Symbol] -> [Symbol]
ersetze_prod a b [] = []
ersetze_prod a b (y:ys) = if b==y 
                             then 
                             a:ersetze_prod a b ys
                             else  
                             y:ersetze_prod a b ys 


--O(n)
split_in_tupel_prod1 :: [Prod]->([Prod],[Prod])
split_in_tupel_prod1 x = split_in_tupel_prod_sub1 x ([],[])


split_in_tupel_prod_sub1:: [Prod]->([Prod],[Prod])->([Prod],[Prod])
split_in_tupel_prod_sub1 []     (prods,lambda_prods) = (prods,lambda_prods)
split_in_tupel_prod_sub1 (x:xs) (prods,lambda_prods) = case x of
                                     (Produktion (p) [N a])     ->  split_in_tupel_prod_sub1 xs (prods,[x]++lambda_prods)
                                     otherwise                  ->  split_in_tupel_prod_sub1 xs ([x]++prods,lambda_prods)
--0 Regel der CNF: Ersetze alle Regeln der Form : N a -> N b und N b -> T x (oder N a -> N b und N b -> N c usw.), durch eine neue Regel N a -> T x (die Regel N a -> N b wird nicht mehr benötigt)




 

































--1.  Regel der CNF: Regeln N a -> T b werden übernommen => snd (split_in_tupel_prod (umwandeln( beispiel1)))  kommt in CNF Menge
--2.  Regel der CNF: Für alle Terminalsymbole aus der Liste  "tprods" (die auf der rechten Seite  alleine stehen), werden in "t_n_prods" eingesetzt.
-------Übergibt der durchlaufe_tprods die Listen nprod, tprod und t_n_prod , die durch split_in_tupel_prod erzeugt werden
cnf_reg_2 :: [Prod] -> ([Prod],[Prod],[Prod])
cnf_reg_2 prod = let stp = (split_in_tupel_prod prod) in
                 durchlaufe_tprods (first stp) (second stp) (thrd stp) []


------------Übergibt alle tprods einzeln an die Funktion setze_in_tnprods_ein
durchlaufe_tprods :: [Prod]->[Prod]->[Prod] -> [Prod] -> ([Prod],[Prod],[Prod])
durchlaufe_tprods nprods [] t_n_prods temp = (nprods,temp,t_n_prods)
durchlaufe_tprods nprods (tprod:ts) t_n_prods temp = let sit = (setze_in_tnprods_ein nprods tprod t_n_prods ([],[],[])) in
                                                     durchlaufe_tprods (first sit) ts (thrd sit) (temp++[tprod]) 

----------Setzt für jedes Terminal das dazugehörige Nichtterminal auf den rechten Seiten aller Produktionen in t_n_prods, in denen das Terminal vorkommt, ein
setze_in_tnprods_ein :: [Prod]->Prod->[Prod]-> ([Prod],[Prod],[Prod]) -> ([Prod],[Prod],[Prod])
setze_in_tnprods_ein  nprod x [] (temp1,temp2,temp3) = (nprod++temp1,temp2,temp3)
setze_in_tnprods_ein  nprod (Produktion (N a) [T b]) ((Produktion (N c) liste):ys) (temp1,temp2,temp3) = 
                   if enthalten (T b) liste 
                   then 
                      let ers = (ersetze1 (Produktion (N a) [T b]) liste) in 
                      if (nurN ers) 
                      then 
                      setze_in_tnprods_ein ([(Produktion (N c) ers)]++nprod) (Produktion (N a) [T b]) ys (temp1,temp2,temp3)
                      else 
                      setze_in_tnprods_ein nprod (Produktion (N a) [T b]) ys (temp1,temp2,[Produktion (N c) ers]++temp3)  
                   else 
                   setze_in_tnprods_ein nprod (Produktion (N a) [T b]) ys (temp1,temp2,[(Produktion (N c) liste)]++temp3)



--O(n)
split_in_tupel_prod :: [Prod]->([Prod],[Prod],[Prod])
split_in_tupel_prod x = split_in_tupel_prodsub (cnf_reg_0 x) ([],[],[])
-----------------------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------
--Bildet ein Tupel mit drei Listen mit (nprods    = "Produktionen nur mit N's aufder rechten Seite",
--                                      tprods    = "Produktionen nur mit einem T auf der rechtren Seiten",
--                                      t_n_prods = "Produktionen mit N's und T's auf der rechten Seite")
--testaufruf  split_in_tupel_prod (beispiel1)--------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------
split_in_tupel_prodsub:: [Prod]->([Prod],[Prod],[Prod])->([Prod],[Prod],[Prod])
split_in_tupel_prodsub []     (nprods,tprods,t_n_prods) = (nprods,tprods,t_n_prods)
split_in_tupel_prodsub (x:xs) (nprods,tprods,t_n_prods) = case x of
                                     (Produktion _ liste)         -> if (nurN liste) 
                                                                       then
                                                                         split_in_tupel_prodsub xs ([x]++nprods,tprods,t_n_prods)
                                                                       else
                                                                        if (length liste)==1
                                                                         then 
                                                                           split_in_tupel_prodsub xs (nprods,[x]++tprods,t_n_prods)
                                                                         else 
                                                                           split_in_tupel_prodsub xs (nprods,tprods,[x]++t_n_prods)



enthalten :: Symbol -> [Symbol] -> Bool
enthalten x []     = False
enthalten x (y:ys) = if x==y then True
                             else enthalten x ys


nurN :: [Symbol] -> Bool
nurN [] = True
nurN (x:xs) = case x of  
                        N a -> nurN xs 
                        otherwise -> False 


ersetze1 :: Prod -> [Symbol] -> [Symbol]
ersetze1 x [] = []
ersetze1 (Produktion (N a) [h]) (y:ys) = if h==y then (N a):ersetze1 (Produktion (N a) [h]) ys
                                                 else  y   :ersetze1 (Produktion (N a) [h]) ys 



--1.  Regel der CNF: Regeln N a -> T b werden übernommen => snd (split_in_tupel_prod (umwandeln( beispiel1)))  kommt in CNF Menge
--2.  Regel der CNF: Für alle Terminalsymbole aus der Liste  "tprods" (die auf der rechten Seite  alleine stehen), werden in "t_n_prods" eingesetzt.




























--2.5 Regel der CNF: Für alle Terminalsymbole, die auf der rechten Seite nicht alleine stehen (t_n_prods) zb. (N a -> Tb, N d ,N c), werden
--Diese Regel benötigt die Ausgabe von cnf_reg_2. Hier muss das dritte Element im Tripel abgefragt werden, weil sich hier eventuell alle Produktionen befinden die Terminalsymbole
--enthalten die noch von keinem Nichtterminal produziert werden.Falls es leer ist, kann mann direkt zu Regel 3 übergehen

--Die Abbruchbedingung übergibt den aktuelsten Zaehlerstand an die funktion cnf_reg_3 für die neuen Nichtterminale die noch erzeugt werden , damit keine schon 
--vorhandenen Nichtterminale (die in cnf_reg_2_5 erzeugt wurden), nochmal verwendet werden    
cnf_reg_2_5 :: ([Prod],[Prod],[Prod]) -> [Prod] -> [Prod] -> Integer -> ([Prod],[Prod],[Prod])
cnf_reg_2_5 (nprod,tprod,[]) [] neuprods zaehler                      =  cnf_reg_3 (nprod,tprod,[]) neuprods zaehler
cnf_reg_2_5 (nprod,tprod,[]) neu_t_n_prods neuprods zaehler           = (cnf_reg_2_5 (nprod,tprod,neu_t_n_prods) [] neuprods zaehler)  ---  
cnf_reg_2_5 (nprod,tprod,(x:t_n_prod)) neu_t_n_prods neuprods zaehler = let v = cnf2_5_sub nprod tprod x neu_t_n_prods neuprods zaehler in
                                                                   cnf_reg_2_5 (first_2 v, second_2 v, t_n_prod) (third_2 v) (fifth_2 v) (frth_2 v) 

cnf2_5_sub :: [Prod] -> [Prod] -> Prod -> [Prod]-> [Prod]-> Integer -> ([Prod],[Prod],[Prod],Integer,[Prod])
cnf2_5_sub nprod tprod (Produktion (N a) liste) neu_t_n_prods neuprods zaehler = let elr = (elementR1 (findeT liste) tprod) 
                                                                                     ers = (ersetzeTN (findeT liste) (snd elr) liste) 
                                                                                     terminal = (Produktion (N (show zaehler)) [(findeT liste)])
                                                                                 in
                                                                          if (fst elr) && (nurN ers)
                                                                          then 
                                                                          ([(Produktion (N a) ers)]++nprod,tprod,neu_t_n_prods,zaehler,neuprods)
                                                                          else 
                                                                          if
                                                                          fst elr
                                                                          then 
                                                                          (nprod,tprod,([(Produktion (N a) ers)]++neu_t_n_prods),zaehler,neuprods)
                                                                          else 
                                                                          (nprod,[terminal]++tprod,([(Produktion (N a) liste)]++neu_t_n_prods),(zaehler+1),[terminal]++neuprods)



first_2 :: ([a],[a],[a],b,[a])->[a]
first_2 (x,y,z,k,l) = x


second_2 :: ([a],[a],[a],b,[a])->[a]
second_2 (x,y,z,k,l) = y


third_2 :: ([a],[a],[a],b,[a])->[a]
third_2 (x,y,z,k,l) = z

frth_2 :: ([a],[a],[a],b,[a])-> b
frth_2 (x,y,z,k,l) = k

fifth_2 :: ([a],[a],[a],b,[a])-> [a]
fifth_2 (x,y,z,k,l) = l

ersetzeTN :: Symbol -> Symbol -> [Symbol] -> [Symbol]
ersetzeTN (T a) (N b) (x:xs) = if x==(T a) then ((N b):xs)
                                           else x: (ersetzeTN (T a) (N b) (xs))

findeT :: [Symbol] -> Symbol
findeT []     = error "Nur N's"
findeT (x:xs) = case x of
              T b       -> T b
              otherwise -> findeT xs

ersetzeNT :: Integer -> Symbol
ersetzeNT zaehler = getStrNT (show zaehler)


getStrNT :: String -> Symbol
getStrNT x = N x

--Es muss überprüft werden ob die neu erzeugten Produktionen nochmal gebraucht werden
elementR1 :: Symbol -> [Prod] -> (Bool,Symbol)
elementR1 x [] = (False,error "Hat keine Produktion")
elementR1 x ((Produktion (N a) [c]):tprod) = if x==c then (True,N a)
                                                     else elementR1 x tprod 



--2.5 Regel der CNF: Für alle Terminalsymbole, die auf der rechten Seite nicht alleine stehen (t_n_prods) zb. (N a -> Tb, N d ,N c), werden
--Diese Regel benötigt die Ausgabe von cnf_reg_2. Hier muss das dritte Element im Tripel abgefragt werden, weil sich hier eventuell alle Produktionen befinden die Terminalsymbole
--enthalten die noch von keinem Nichtterminal produziert werden.Falls es leer ist, kann mann direkt zu Regel 3 übergehen

























--3.  Regel der CNF: Die restlichen Regeln haben die Form N a -> N y1....N yn. Wenn n/=2 werden weitere Nichtterminale eingeführt, sodass N a -> N y1 N z1 bis N zn-2 -> N yn-1 N yn
cnf_reg_3 :: ([Prod],[Prod],[Prod]) -> [Prod] -> Integer -> ([Prod],[Prod],[Prod])
cnf_reg_3 ([],tprods,fertigenprods) neuprods zaehler = (fertigenprods,tprods,neuprods)
cnf_reg_3 (((Produktion (N a) xs ):nprods),tprods,fertigenprods) neuprods zaehler = if ((length xs)==2) 
                                                                                    then 
                                                                                    cnf_reg_3 (nprods,tprods,[(Produktion (N a) xs )]++fertigenprods) neuprods zaehler
                                                                                    else 
                                                                                    if (length xs)<2 then error "Fehler bei cnf_reg_0" else
                                                                                    let rb  = (runterbrechen zaehler (Produktion (N a) xs) fertigenprods neuprods)
                                                                                        bir = (first3 rb)
                                                                                        iki = (second3 rb)
                                                                                        uec = (thrd3 rb) in
                                                                                      cnf_reg_3 (nprods,tprods,bir) uec iki


first3 :: ([a],b,[a])->[a]
first3 (x,y,z) = x


second3 :: ([a],b,[a])->b
second3 (x,y,z) = y


thrd3 :: ([a],b,[a])->[a]
thrd3 (x,y,z) = z

---Splitet für eine Produktion die rechte Seite in mehrere neue Produktionen der länge 2 auf. 
---Falls schon Produktionen existieren werden diese übernommen (N 1 -> N a N b , dann wird keine neue Produktion N 2 -> N a N b hinzugefügt,
---sonder N 1 wird übernommen), ohne zu erzeugen  
runterbrechen :: Integer -> Prod -> [Prod] -> [Prod] -> ([Prod],Integer,[Prod]) 
runterbrechen zaehler (Produktion (N a) (x:xs)) fuereinprodfertig neuprods = let isent     = (istenthalten xs fuereinprodfertig)
                                                                                 isent2    = (istenthalten ys fuereinprodfertig)
                                                                                 ys        = [x,(head xs)]
                                                                                 len       = (length xs)==2 
                                                                                 chomskyNT = [(Produktion (N a) [x, (N(show zaehler))]),(Produktion (N(show zaehler)) xs)]
                                                                                 prod1     = (Produktion (N a) [x, (getProd1 xs fuereinprodfertig)])
                                                                                 prod2     = (Produktion (N a) [(getProd1 ys fuereinprodfertig),(last xs)])
                                                                                 prod3     = (Produktion (N a) [x,(getProd1 xs fuereinprodfertig)])
                                                                                 prod4     = (Produktion (N a) [x,(N(show zaehler))])
                                                                          in
                                                                          if len && isent
                                                                          then 
                                                                          ([prod1]++fuereinprodfertig,zaehler,[prod1]++neuprods)
                                                                          else 
                                                                          if len && isent2
                                                                          then 
                                                                          ([prod2]++fuereinprodfertig,zaehler,[prod2]++neuprods)
                                                                          else 
                                                                          if len
                                                                          then
                                                                          (chomskyNT++fuereinprodfertig,(zaehler+1),chomskyNT++neuprods)
                                                                          else 
                                                                          if isent
                                                                          then
                                                                          runterbrechen zaehler (Produktion (N a) xs) ([prod3]++fuereinprodfertig) ([prod3]++neuprods)
                                                                          else 
                                                                          runterbrechen (zaehler+1)(Produktion (N(show zaehler)) xs)([prod4]++fuereinprodfertig)([prod4]++neuprods)

istenthalten :: [Symbol] -> [Prod] -> Bool
istenthalten x [] = False
istenthalten x ((Produktion (N a) liste):xs) = if x==liste then True
                                                           else istenthalten x xs

getProd1 :: [Symbol] -> [Prod] -> Symbol
getProd1 x ((Produktion (N a) liste):xs) = if x==liste then N a
                                                       else getProd1 x xs

--3.  Regel der CNF: Die restlichen Regeln haben die Form N a -> N y1....N yn. Wenn n/=2 werden weitere Nichtterminale eingeführt, sodass N a -> N y1 N z1 bis N zn-2 -> N yn-1 N yn
