module PlandowskiSTG_Prod.PlandowskiSTG_Prod where

import Datentypen
import PlandowskiSTG_Prod.Plandowski2 (plandowski)
import PlandowskiSTG_Prod.ChomskyProd (chomsky)
import Beispiele
import Beispiele_neu
import Debug.Trace

----plandowskiSTG_Prod beispiel [(TN "A_s",TN "A_t")]
----plandowskiSTG_Prod beispiel_18 [(TN "A_t",TN "A_s")]
plandowskiSTG_Prod:: [STG_Prod] -> [(STG_Pre_Symbol,STG_Pre_Symbol)] -> Bool
plandowskiSTG_Prod stgprods set = let
                                     prod    = (stg_in_prod stgprods)
                                     chomsk  = (chomsky prod 1)
                                     testset = (presym_in_sym set)
                                  in
 
                                     plandowski chomsk testset




















presym_in_sym :: [(STG_Pre_Symbol,STG_Pre_Symbol)] -> [(Symbol,Symbol)]
presym_in_sym []     = []
presym_in_sym (x:xs) = case x of
                            ((TN a),(TN b)) -> (N ("TN-"++a),N ("TN-"++b)) : presym_in_sym xs
                            ((TN a),(CN b)) -> (N ("TN-"++a),N ("CN-"++b)) : presym_in_sym xs
                            ((CN a),(CN b)) -> (N ("CN-"++a),N ("CN-"++b)) : presym_in_sym xs
                            ((CN a),(TN b)) -> (N ("CN-"++a),N ("TN-"++b)) : presym_in_sym xs



lochReduktion :: [STG_Prod] -> [STG_Prod]
lochReduktion prod = fst (durchlaufeLoch (split_in_tupel prod))

 
durchlaufeLoch :: ([STG_Prod],[STG_Prod]) -> ([STG_Prod],[STG_Prod])
durchlaufeLoch ((prods),[]) = (prods,[])
durchlaufeLoch ((prods),(lambda:ys))  = durchlaufeLoch (durchlaufeProd prods lambda  ([],ys))


durchlaufeProd :: [STG_Prod]-> STG_Prod->([STG_Prod],[STG_Prod])->([STG_Prod],[STG_Prod])
durchlaufeProd [] lambda (neuprods,neulambda)    = (neuprods,neulambda)
durchlaufeProd (prod:ys) lambda (neuprods,neulambda) =  let entp = entferneProd prod lambda ([],[]) in
                                                        durchlaufeProd ys lambda ((fst entp)++neuprods,neulambda++(snd entp))


entferneProd :: STG_Prod-> STG_Prod -> ([STG_Prod],[STG_Prod])-> ([STG_Prod],[STG_Prod])
entferneProd (STG_Prod p (ks)) (STG_Prod (TN b) _) (neuprods,neulambda) = let flrb = (filter (/=(TN b)) ks) in
                                                                                if (flrb /= [])
                                                                                then
                                                                                ([(STG_Prod p flrb)]++neuprods,neulambda)
                                                                                else
                                                                                (neuprods,[(STG_Prod p flrb)]++neulambda)

entferneProd (STG_Prod p (ks)) (STG_Prod (CN b) _) (neuprods,neulambda) = let fllb = (filter (/=(CN b)) ks) in
                                                                                if (fllb/=[])
                                                                                then
                                                                                ([(STG_Prod p fllb)]++neuprods,neulambda)
                                                                                else
                                                                                (neuprods,[(STG_Prod p fllb)]++neulambda)



split_in_tupel :: [STG_Prod]->([STG_Prod],[STG_Prod])
split_in_tupel x = split_in_tupel_sub x ([],[])


split_in_tupel_sub:: [STG_Prod]->([STG_Prod],[STG_Prod])->([STG_Prod],[STG_Prod])
split_in_tupel_sub []     (temp,temp2) = (temp,temp2)
split_in_tupel_sub (x:xs) (temp,temp2) = case x of
                                     (STG_Prod _ [Loch])      ->  split_in_tupel_sub xs (temp,[x]++temp2)
                                     otherwise                ->  split_in_tupel_sub xs ([x]++temp,temp2)



stg_in_prod :: [STG_Prod]->[Prod]
stg_in_prod x = umwandeln_sub (lochReduktion x)

umwandeln_sub::[STG_Prod]->[Prod]
umwandeln_sub []                          = []
umwandeln_sub ((STG_Prod y list):xs)  = case y of 
                                             (TN a)         ->   Produktion (N ("TN-"++a)) (umwandeln1 list):umwandeln_sub xs
                                             (CN a)         ->   Produktion (N ("CN-"++a)) (umwandeln1 list):umwandeln_sub xs
                                             otherwise      ->   error "unerlaubte Produktion"

umwandeln1 :: [STG_Pre_Symbol] -> [Symbol]
umwandeln1 [] = []
umwandeln1 (x:xs) = case x of
                           CN a      ->  (N ("CN-"++a)) : umwandeln1 xs
                           TN a      ->  (N ("TN-"++a)) : umwandeln1 xs
                           F a       ->  (T ("F-"++a)) : umwandeln1 xs
                           X a       ->  (T ("X-"++a)) : umwandeln1 xs
                           K a       ->  (T ("K-"++a)) : umwandeln1 xs
                           otherwise -> error "unerlaubter Typ auf der rechten Seite (Module PlandowskiSTG, Funktion umwandeln1)"













