module STG_in_SCFG.STG_in_SCFG where

import Datentypen
import Data.List
import Debug.Trace



----------------------------------------------------------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-------------------------------------------- Transformiert die STG ind die Preorder Traversierung (Regel Fig.1)-------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------------------------------------------------------
transform_p :: [STG_Prod] -> [ProdPre]
transform_p x = transform_p_sub x []

--Nach Fig1. erzeugt Die SCFG Preorder Traversal einer STG Grammatik . O(n*5n)=O(n^2)
transform_p_sub :: [STG_Prod] -> [ProdPre] -> [ProdPre]
transform_p_sub [] temp = temp 
transform_p_sub (x:xs) temp = case x of 
{-1.Regel-}                    STG_Prod (TN a) (F f : ys)    -> if nurTN ys then 
                                                                            transform_p_sub xs ([ProdPreOrder (P a)  ( [F f]++convert ys )]++temp) 
                                                                            else 
                                                                            error "1. Keine STG, Module STG_in_SCFG " 
{-2.Regel-}                    STG_Prod (TN a) [CN b, TN c]  -> transform_p_sub xs ([ProdPreOrder (P a) [L b, P c, R b]]++temp)
{-3.Regel-}                    STG_Prod (TN a) [TN b]        -> transform_p_sub xs ([ProdPreOrder (P a) [P b]]++temp)
{-4.Regel-}                    STG_Prod (CN c) [CN c1 ,CN c2]-> transform_p_sub xs ([ProdPreOrder (L c) [L c1, L c2], ProdPreOrder (R c) [R c2, R c1]]++temp)
{-5.Regel-}                    STG_Prod (CN a) [F f, CN b]   -> transform_p_sub xs ([ProdPreOrder (L a) [F f, L b], ProdPreOrder (R a) [R b]]++temp) 
                               STG_Prod (CN a) [F f, TN b]   -> error ("Nicht erlaubte Produktion : CN "++show a++"-->"++"[F "++show f++", TN "++show b++" ! Auf der rechten Seite kommt kein Kontext vor!")
                               STG_Prod (CN a) (F f:ys)      ->  if (length.fst.fst.einCN $ ys) == 1 
                                                                 then 
                                                                 transform_p_sub xs ([ProdPreOrder (L a) ([F f]++ (fst.snd.einCN $ ys)),ProdPreOrder (R a) (snd.snd.einCN $ ys)]++temp) 
                                                                 else 
                                                                 error ("Auf der rechten Seite vom Kontext CN "++show a++"tauchen entweder nur TNs auf,oder es taucht kein Kontext auf, oder es taucht mehr als ein Kontext auf!")
{-6.Regel-}                    STG_Prod (CN c) [Loch]        -> transform_p_sub xs ([ProdPreOrder (L c) [Lambda], ProdPreOrder (R c) [Lambda]]++temp)
{-7.Regel-}                    STG_Prod (TN a) [X x]         -> transform_p_sub xs ([ProdPreOrder (P a) [X x]]++temp)
{-8.Regel-}                    STG_Prod (TN a) [K k]         -> transform_p_sub xs ([ProdPreOrder (P a) [K k]]++temp)
                               STG_Prod (a)    [F f]         -> error "Funktion ohne Argument ist nicht erlaubt"
                               STG_Prod (a)    [CN b]        -> error "Kontext darf nicht alleine stehen"
                               otherwise                     -> error "2. Keine STG, Module STG_in_SCFG "



--Überprüft ob die rechte Seite nur Term-Nonterminale enthält O(n)
nurTN :: [STG_Pre_Symbol] -> Bool
nurTN [] = True
nurTN (x:xs) = case x of  
                TN a -> nurTN xs 
                otherwise -> False 

--konvertiert für die erste Regel ((A -> f(A_1,.....,A_m) => P_A -> f PA_1...,PA_m)) die Stg Term-Non-Terminale in Preorder NonTErminale -- O(n)  
convert :: [STG_Pre_Symbol]  -> [STG_Pre_Symbol]
convert []  = []
convert (TN a:xs)  =  (P a) : convert xs

--Hilfsfunktion O(1)
getStr :: [STG_Pre_Symbol] -> String
getStr [] = "Fehler , Module STG_in_SCFG , Funktion getStr"
getStr [CN x] = x
getStr [TN x] = x


--O(n)
-- Erzeugt die Regel C -> f(A_1,.....,A_i-1,C_i,A_i+1,....A_m) => L_C -> f PA_1...PA_i-1 LC_i , R_C -> RC_i,PA_i+1,...PA_m
-- tempL = L_C , tempR = R_C, temp2 = PA_1...PA_i-1 LC_i , temp3 = RC_i,PA_i+1,...PA_m
-- Solange kein neues tempR ( if (length tempR == 0)) dazukommt wird die linke Seite erzeugt, (tempL,tempR,(((((temp2++[P a])))))),temp3)
-- Ansonsten ist das  erste CN gekommen und die Rechte Seite wird erzeugt (tempL,tempR,temp2++tempL,((((((((((((temp3++tempR++[P a])))))))))))))
einCN :: [STG_Pre_Symbol] -> (([STG_Pre_Symbol],[STG_Pre_Symbol]),([STG_Pre_Symbol],[STG_Pre_Symbol]))
einCN xs  = einCN_sub xs ([],[],[],[]) 

einCN_sub :: [STG_Pre_Symbol] -> ([STG_Pre_Symbol],[STG_Pre_Symbol],[STG_Pre_Symbol],[STG_Pre_Symbol]) -> (([STG_Pre_Symbol],[STG_Pre_Symbol]),([STG_Pre_Symbol],[STG_Pre_Symbol]))
einCN_sub []     (tempL,tempR,temp2,temp3) = ((tempL,tempR),(temp2,temp3))
einCN_sub (x:xs) (tempL,tempR,temp2,temp3) = case x of 
                                          TN a -> if (length tempR == 0) then einCN_sub xs (tempL,tempR,temp2++[P a],temp3)
                                                                         else einCN_sub xs (tempL,tempR,temp2,temp3++[P a])
                                          CN a -> einCN_sub xs ([L a]++tempL,[R a]++tempR,temp2++[L a],[R a]++temp3) 

--Hilfsfunktion O(1) 
liste_in_tupel:: [STG_Pre_Symbol] -> (STG_Pre_Symbol)
liste_in_tupel (x:xs) = (x)
 
----------------------------------------------------------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-------------------------------------------- Transformiert die STG ind die Preorder Traversierung (Regel Fig.1)-------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------------------------------------------------------------------
